import logging
from ..common.common import SectionHandler


class References(SectionHandler):
    def __init__(self, config):
        super().__init__()
        self.force_default_category = config.get('force_insert_default_reference_category')

    def _process_mandatory_elements(self, root_element):

        references = []

        for reference in root_element.Reference:
            ref_csaf = {'summary': reference.Description.text,
                        'url': reference.URL.text}

            if reference.attrib.get('Type'):
                ref_csaf['category'] = reference.attrib['Type'].lower()
            elif self.force_default_category:
                ref_csaf['category'] = 'external'
                logging.info('"Type" attribute not present in "Reference" element, using default value "external". '
                             'This can be controlled by "force_insert_default_reference_category" option.')

            references.append(ref_csaf)

        self.csaf = references

    def _process_optional_elements(self, root_element):
        pass



import logging
import re
import sys
from operator import itemgetter
from typing import Tuple
from ..common.common import SectionHandler
from ..common.utils import get_utc_timestamp


class DocumentTracking(SectionHandler):
    tracking_status_mapping = {
        'Final': 'final',
        'Draft': 'draft',
        'Interim': 'interim',
    }

    def __init__(self, config, pkg_version):
        super().__init__()
        self.cvrf2csaf_name = config.get('cvrf2csaf_name')
        self.cvrf2csaf_version = pkg_version
        self.fix_insert_current_version_into_revision_history \
            = config.get('fix_insert_current_version_into_revision_history')

    def _process_mandatory_elements(self, root_element):
        self.csaf['id'] = root_element.Identification.ID.text
        self.csaf['current_release_date'] = get_utc_timestamp(root_element.CurrentReleaseDate.text)
        self.csaf['initial_release_date'] = get_utc_timestamp(root_element.InitialReleaseDate.text)
        self.csaf['status'] = self.tracking_status_mapping[root_element.Status.text]

        revision_history, version = self._handle_revision_history_and_version(root_element)
        self.csaf['revision_history'] = revision_history
        self.csaf['version'] = version

        # Generator is set by this Converter
        self.csaf['generator'] = {}
        self.csaf['generator']['date'] = get_utc_timestamp(time_stamp='now')
        self.csaf['generator']['engine'] = {}
        self.csaf['generator']['engine']['name'] = self.cvrf2csaf_name
        self.csaf['generator']['engine']['version'] = self.cvrf2csaf_version

    def _process_optional_elements(self, root_element):
        if hasattr(root_element.Identification, 'Alias'):
            aliases = []
            for alias in root_element.Identification.Alias:
                aliases.append(alias.text)

            self.csaf['aliases'] = aliases

    @staticmethod
    def check_for_version_t(revision_history):
        """
        Checks whether all version numbers in /document/tracking/revision_history match semantic versioning.
        semantic version is defined in version_t definition
        see: https://docs.oasis-open.org/csaf/csaf/v2.0/csd01/csaf-v2.0-csd01.html#3111-version-type and
        section 9.1.5 Conformance Clause 5: CVRF CSAF converter
        """

        pattern = (
            r'^((0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)'
            r'(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?)$'
        )
        return all(re.match(pattern, revision['number']) for revision in revision_history)

    @staticmethod
    def _as_int_tuple(text: str) -> Tuple[int]:
        """ Converts string of dotted numbers into tuple of ints """
        try:
            return tuple(int(part) for part in text.split('.'))
        except ValueError:
            return (sys.maxsize, )

    def _add_current_revision_to_history(self, root_element, revision_history) -> None:
        """
        If the current version is missing in Revision history and --fix-insert-current-version-into-revision-history is True,
        the current version is added to the history.
        """

        revision_history.append(
            {
                'date': get_utc_timestamp(root_element.CurrentReleaseDate.text),
                'number': root_element.Version.text,
                'summary': f'Added by {self.cvrf2csaf_name} as the value was missing in the original CVRF.',
                # Extra vars
                'number_cvrf': root_element.Version.text,
                'version_as_int_tuple': self._as_int_tuple(root_element.Version.text),
            }
        )

    @staticmethod
    def _reindex_versions_to_integers(root_element, revision_history):
        logging.warning('Some version numbers in revision_history do not match semantic versioning. '
                        'Reindexing to integers.')

        revision_history_sorted = sorted(revision_history, key=itemgetter('version_as_int_tuple'))

        for rev_number, revision in enumerate(revision_history_sorted, start=1):
            revision['number'] = str(rev_number)

        # after reindexing, match document version to corresponding one in revision history
        version = next(rev for rev in revision_history_sorted if rev['number_cvrf'] == root_element.Version.text)['number']

        return revision_history_sorted, version

    def _handle_revision_history_and_version(self, root_element):
        # preprocess the data
        revision_history = []
        for revision in root_element.RevisionHistory.Revision:
            # number_cvrf: keep original value in this variable for matching later
            # number: this value might be overwritten later if some version numbers doesn't match semantic versioning
            revision_history.append(
                {
                    'date': get_utc_timestamp(revision.Date.text),
                    'number': revision.Number.text,
                    'summary': revision.Description.text,
                    # Extra vars
                    'number_cvrf': revision.Number.text,
                    'version_as_int_tuple': self._as_int_tuple(revision.Number.text),
                }
            )

        # Just copy over the version
        version = root_element.Version.text

        missing_latest_version_in_history = False
        # Do we miss the current version in the revision history?
        if not [rev for rev in revision_history if rev['number'] == version]:
            if self.fix_insert_current_version_into_revision_history:
                logging.warning('Trying to fix the revision history by adding the current version. '
                                'This may lead to inconsistent history. This happens because '
                                '--fix-insert-current-version-into-revision-history is used. ')
                self._add_current_revision_to_history(root_element, revision_history)
            else:
                logging.error('Current version is missing in revision history. '
                              'This can be fixed by using --fix-insert-current-version-into-revision-history')
                missing_latest_version_in_history = True
                self.error_occurred = True

        # handle corresponding part of Conformance Clause 5: CVRF CSAF converter
        # that is: some version numbers in revision_history don't match semantic versioning
        if not self.check_for_version_t(revision_history):
            if not missing_latest_version_in_history:
                revision_history, version = self._reindex_versions_to_integers(root_element, revision_history)
            else:
                logging.error('Can not reindex revision history to integers because of missing the current version. '
                              'This can be fixed with --fix-insert-current-version-into-revision-history')
                self.error_occurred = True

        # cleanup extra vars
        for revision in revision_history:
            revision.pop('number_cvrf')
            revision.pop('version_as_int_tuple')

        return revision_history, version



import logging
from ..common.common import SectionHandler


class Acknowledgments(SectionHandler):
    def __init__(self):
        super().__init__()

    def _process_mandatory_elements(self, root_element):

        if hasattr(root_element, 'Acknowledgment'):
            # at least one entry in list of Acknowledgments
            pass # No field is mandatory due to CVRF.xsd; version 1.2


    def _process_optional_elements(self, root_element):
        self.csaf = []

        for ack in root_element.Acknowledgment:
            # empty Acknowledgment slips CVRF input validation
            if not any([hasattr(ack, 'Name'),
                        hasattr(ack, 'Organization'),
                        hasattr(ack, 'Description'),
                        hasattr(ack, 'URL')]):
                logging.warning(f'Skipping empty Acknowledgment entry, input line: {ack.sourceline}')
                continue

            ack_elem = {}

            if hasattr(ack, 'Organization'):
                if len(ack.Organization) > 1:
                    # If more than one cvrf:Organization instance is given,
                    # the CVRF CSAF converter converts the first one into the organization.
                    # In addition the converter outputs a warning that information might be lost during conversion
                    # of document or vulnerability acknowledgment.
                    logging.warning(f'CSAF 2.0 allows only one organization inside Acknowledgments. '
                                    f'Taking the first occurence, ignoring: {ack.Organization[1:]}.')

                ack_elem['organization'] = ack.Organization[0].text

            if hasattr(ack, 'Description'):
                # Single Description elem is asserted on the input
                ack_elem['summary'] = ack.Description[0].text

            # Names and URLs can have more entries
            if hasattr(ack, 'Name'):
                ack_elem['names'] = [x.text for x in ack.Name]

            if hasattr(ack, 'URL'):
                ack_elem['urls'] = [x.text for x in ack.URL]

            self.csaf.append(ack_elem)



import logging
import bisect
import re

from collections import defaultdict
from itertools import chain

from ..common.common import SectionHandler
from ..section_handlers.document_acknowledgments import Acknowledgments
from ..section_handlers.references import References
from ..section_handlers.notes import Notes
from ..common.utils import get_utc_timestamp


class Vulnerability(SectionHandler):
    # mapping of CVRF XML elements to CSAF json keys
    cvss_v3_mapping = {'BaseScoreV3': 'baseScore',
                       'TemporalScoreV3': 'temporalScore',
                       'EnvironmentalScoreV3': 'environmentalScore',
                       'VectorV3': 'vectorString'}

    cvss_v2_mapping = {'BaseScoreV2': 'baseScore',
                       'TemporalScoreV2': 'temporalScore',
                       'EnvironmentalScoreV2': 'environmentalScore',
                       'VectorV2': 'vectorString'}

    def __init__(self, config):
        super().__init__()
        self.csaf = []
        self.config = config
        self.remove_CVSS_values_without_vector = config['remove_CVSS_values_without_vector']
        self.default_CVSS3_version = config['default_CVSS3_version']

    def _process_mandatory_elements(self, root_element):
        pass

    @staticmethod
    def _handle_involvements(root_element):
        involvements = []
        for involvement_elem in root_element.Involvement:
            involvement = {'party': involvement_elem.attrib['Party'].lower(),
                           'status': involvement_elem.attrib['Status'].lower().replace(' ', '_')}

            if hasattr(involvement_elem, 'Description'):
                involvement['summary'] = involvement_elem.Description.text
            involvements.append(involvement)

        return involvements

    @staticmethod
    def _handle_product_statuses(root_element):
        statuses = defaultdict(list)
        for status_elem in root_element.Status:
            status_type = status_elem.attrib['Type'].lower().replace(' ', '_')
            product_ids = [product_id.text for product_id in status_elem.ProductID]
            statuses[status_type].extend(product_ids)

        return statuses

    @staticmethod
    def _handle_threats(root_element):
        threats = []
        for threat_elem in root_element.Threat:
            threat = {'details': threat_elem.Description.text,
                      'category': threat_elem.attrib['Type'].lower().replace(' ', '_')}

            if hasattr(threat_elem, 'ProductID'):
                threat['product_ids'] = [product_id.text for product_id in threat_elem.ProductID]

            if hasattr(threat_elem, 'GroupID'):
                threat['group_ids'] = [group_id.text for group_id in threat_elem.GroupID]

            if 'Date' in threat_elem.attrib:
                threat['date'] = get_utc_timestamp(threat_elem.attrib['Date'])

            threats.append(threat)

        return threats

    @staticmethod
    def _handle_remediations(root_element):

        category_mapping = {'Workaround': 'workaround',
                            'Mitigation': 'mitigation',
                            'Vendor Fix': 'vendor_fix',
                            'None Available': 'none_available',
                            'Will Not Fix': 'no_fix_planned'}

        remediations = []
        for remediation_elem in root_element.Remediation:
            remediation = {'category': category_mapping[remediation_elem.attrib['Type']],
                           'details': remediation_elem.Description.text}

            if hasattr(remediation_elem, 'Entitlement'):
                remediation['entitlements'] = [entitlement.text for entitlement in remediation_elem.Entitlement]

            if hasattr(remediation_elem, 'URL'):
                remediation['url'] = remediation_elem.URL.text

            if hasattr(remediation_elem, 'ProductID'):
                remediation['product_ids'] = [product_id.text for product_id in remediation_elem.ProductID]

            if hasattr(remediation_elem, 'GroupID'):
                remediation['group_ids'] = [group_id.text for group_id in remediation_elem.GroupID]

            if 'Date' in remediation_elem.attrib:
                remediation['date'] = get_utc_timestamp(remediation_elem.attrib['Date'])

            remediations.append(remediation)

        return remediations

    @staticmethod
    def _base_score_to_severity(base_score):
        base_severity = ((0, 'NONE'), (3.9, 'LOW'), (6.9, 'MEDIUM'), (8.9, 'HIGH'), (10, 'CRITICAL'))
        return base_severity[bisect.bisect_right(base_severity, (base_score, ''))][1]

    def _parse_score_set(self, score_set_element, mapping, version, json_property, product_status):
        """Parses ScoreSetV2 or ScoreSetV3 element."""

        # Parse all input elements except ProductID
        # note: baseScore is always present since it's mandatory for the input
        cvss_score = {csaf: getattr(score_set_element, cvrf).text for cvrf, csaf in mapping.items() if hasattr(score_set_element, cvrf)}

        # Convert all possible scores to float
        scores = ['baseScore', 'temporalScore', 'environmentalScore']
        for score in scores:
            if cvss_score.get(score):
                cvss_score[score] = float(cvss_score[score])

        # Only cvss_v3 has baseSeverity
        if json_property == 'cvss_v3':
            cvss_score['baseSeverity'] = self._base_score_to_severity(cvss_score['baseScore'])

        # HANDLE product_ids
        product_ids = []
        # if we have ProductID element(s), parse and use
        if hasattr(score_set_element, 'ProductID'):
            product_ids = [product_id.text for product_id in score_set_element.ProductID]
        # one of the conversion rules specifies what to do in case of missing ProductID element
        elif product_status:
            states = ('known_affected', 'first_affected', 'last_affected')
            product_ids = list(set(product_id for state in states for product_id in product_status.get(state, [])))

        if len(product_ids) == 0:
            SectionHandler.error_occurred = True
            logging.error('No product_id entry for CVSS score set.')

        # HANDLE vectorString
        # if missing, conversion fails unless remove_CVSS_values_without_vector is true
        # if remove_CVSS_values_without_vector is true, we just ignore the score_set
        if 'vectorString' not in cvss_score:
            if self.remove_CVSS_values_without_vector:
                logging.warning('No CVSS vector string found on the input, ignoring ScoreSet element due to'
                                '"remove_CVSS_values_without_vector" option.')
                return None
            else:
                SectionHandler.error_occurred = True
                logging.error('No CVSS vector string found on the input.')

        # DETERMINE CVSS v 3.x from namespace
        cvss_3_regex = '.*cvss-v(3\.[01]).*'
        match = re.match(cvss_3_regex, score_set_element.tag)
        if match:
            version = match.groups()[0]

        # DETERMINE CVSS v 3.x from vector if present
        if 'vectorString' in cvss_score and json_property == 'cvss_v3':
            # Regex for determining the CVSS version
            regex = "CVSS:(3\.[01]).*"
            match = re.match(regex, cvss_score['vectorString'])
            if not match:
                SectionHandler.error_occurred = True
                logging.error(f'CVSS vector {cvss_score["vectorString"]} is not valid.')
            else:
                version = match.groups()[0]

        cvss_score['version'] = version

        score = {json_property: cvss_score,
                 'products': product_ids}

        return score

    @staticmethod
    def _remove_cvssv3_duplicates(scores):
        """
        Removes products/cvssv3.x score sets for products having both v3.0 and v3.1 score. Three step approach:
         - find products having both versions specified
         - remove those products from score set with version 3.0
         - removes score sets with no products
        """
        # STEP 1
        products_v3_1 = set(chain.from_iterable([score_set['products'] for score_set in scores if
                                                 'cvss_v3' in score_set and score_set['cvss_v3']['version'] == '3.1']))
        products_v3_0 = set(chain.from_iterable([score_set['products'] for score_set in scores if
                                                 'cvss_v3' in score_set and score_set['cvss_v3']['version'] == '3.0']))

        both_versions = products_v3_0.intersection(products_v3_1)

        # STEP 2
        for score_set in scores:
            if 'cvss_v3' in score_set and score_set['cvss_v3']['version'] == '3.0':
                score_set['products'] = [product for product in score_set['products'] if product not in both_versions]

        # STEP 3
        scores = [score_set for score_set in scores if len(score_set['products']) > 0]

        return scores


    def _handle_scores(self, root_element, product_status):
        scores = []

        score_variants = (
            ('ScoreSetV2', self.cvss_v2_mapping, '2.0', 'cvss_v2'),
            ('ScoreSetV3', self.cvss_v3_mapping, self.default_CVSS3_version, 'cvss_v3'),
        )
        for score_variant, mapping, score_version, target in score_variants:
            for score_set in root_element.findall(f'{{*}}{score_variant}'):
                score = self._parse_score_set(score_set, mapping, score_version, target, product_status)
                if score is not None:
                    scores.append(score)

        scores = self._remove_cvssv3_duplicates(scores)

        return scores

    def _process_optional_elements(self, root_element):
        vulnerability = {}

        if hasattr(root_element, 'Acknowledgments'):
            # reuse Acknowledgments handler
            acknowledgments_handler = Acknowledgments()
            acknowledgments_handler.create_csaf(root_element.Acknowledgments)
            vulnerability['acknowledgments'] = acknowledgments_handler.csaf

        if hasattr(root_element, 'CVE'):
            # "^CVE-[0-9]{4}-[0-9]{4,}$" differs from the CVRF regex. Will be checked by json schema validation.
            vulnerability['cve'] = root_element.CVE.text

        if hasattr(root_element, 'CWE'):
            CWE_elements = root_element.CWE
            if len(CWE_elements) > 1:
                logging.warning(f'{len(CWE_elements)} CWE elements found, using only the first one.')
            vulnerability['cwe'] = {'id': CWE_elements[0].attrib['ID'],
                                    'name': CWE_elements[0].text}

        if hasattr(root_element, 'DiscoveryDate'):
            vulnerability['discovery_date'] = get_utc_timestamp(root_element.DiscoveryDate.text)

        if hasattr(root_element, 'ID'):
            vulnerability['id'] = {'system_name': root_element.ID.attrib['SystemName'],
                                   'text': root_element.ID.text}

        if hasattr(root_element, 'Involvements'):
            vulnerability['involvements'] = self._handle_involvements(root_element.Involvements)

        if hasattr(root_element, 'Notes'):
            # reuse Notes handler
            notes_handler = Notes()
            notes_handler.create_csaf(root_element.Notes)
            vulnerability['notes'] = notes_handler.csaf

        if hasattr(root_element, 'ProductStatuses'):
            vulnerability['product_status'] = self._handle_product_statuses(root_element.ProductStatuses)

        if hasattr(root_element, 'References'):
            # reuse References handler
            references_handler = References(config=self.config)
            references_handler.create_csaf(root_element.References)
            vulnerability['references'] = references_handler.csaf

        if hasattr(root_element, 'ReleaseDate'):
            vulnerability['release_date'] = get_utc_timestamp(root_element.ReleaseDate.text)

        if hasattr(root_element, 'Remediations'):
            vulnerability['remediations'] = self._handle_remediations(root_element.Remediations)

        if hasattr(root_element, 'CVSSScoreSets'):
            vulnerability['scores'] = self._handle_scores(root_element.CVSSScoreSets,
                                                          vulnerability.get('product_status'))

        if hasattr(root_element, 'Threats'):
            vulnerability['threats'] = self._handle_threats(root_element.Threats)

        if hasattr(root_element, 'Title'):
            vulnerability['title'] = root_element.Title.text

        self.csaf.append(vulnerability)




from ..common.common import SectionHandler


class DocumentPublisher(SectionHandler):
    type_category_mapping = {
        'Vendor': 'vendor',
        'Coordinator': 'coordinator',
        'User': 'user',
        'Discoverer': 'discoverer',
        'Other': 'other',
    }

    def __init__(self, config):
        super().__init__()
        self.name = config.get('publisher_name')
        self.namespace = config.get('publisher_namespace')

    def _process_mandatory_elements(self, root_element):
        self.csaf['name'] = self.name
        self.csaf['namespace'] = self.namespace
        self.csaf['category'] = self.type_category_mapping[root_element.attrib['Type']]

    def _process_optional_elements(self, root_element):
        # optional values
        if hasattr(root_element, 'ContactDetails'):
            self.csaf['contact_details'] = root_element.ContactDetails.text
        if hasattr(root_element, 'IssuingAuthority'):
            self.csaf['issuing_authority'] = root_element.IssuingAuthority.text




import json
import logging
import os
import re
import yaml
import pkg_resources

from pathlib import Path
from datetime import datetime, timezone
from .common import SectionHandler


def critical_exit(msg, status_code=1):
    """ A critical error encountered, converter is not able to proceed and exits with a status code (default 1) """
    logging.critical(msg)
    exit(status_code)


def handle_boolean_config_values(key, val):
    try:
        if isinstance(val, bool):
            return val
        if val.strip().lower() in {'true', 'yes', '1', 'y'}:
            return True
        if val.strip().lower() in {'false', 'no', '0', 'n'}:
            return False

        critical_exit(f"Reading config.yaml failed. "
                      f"Invalid value for config key {key}: {val}. {e}.")

    except Exception as e:
        critical_exit(f"Reading config.yaml failed. "
                      f"Invalid value for config key {key}: {val} {e}.")


def get_config_from_file() -> dict:
    """ Loads configuration file. Parts of it can be overwritten by CLI arguments. """
    config = dict()
    try:
        # TODO: Workaround for now, config file placement is to be discussed
        req = pkg_resources.Requirement.parse('cvrf2csaf')
        path_to_conf = pkg_resources.resource_filename(req, 'cvrf2csaf/config/config.yaml')
        with open(path_to_conf, 'r') as f:
            config = yaml.safe_load(f)

        for key in ['force', 'fix_insert_current_version_into_revision_history']:
            if key in config.keys():
                config[key] = handle_boolean_config_values(key=key, val=config[key])

    except Exception as e:
        critical_exit(f"Reading config.yaml failed: {e}.")
    finally:
        return config


def create_file_name(document_tracking_id, valid_output):
    if document_tracking_id is not None:
        file_name = re.sub(r"([^+\-_a-z0-9]+)", '_', document_tracking_id.lower())
    else:
        file_name = 'out'

    if not valid_output:
        file_name = f'{file_name}_invalid'
    file_name = f'{file_name}.json'
    return file_name


def store_json(js, fpath):
    try:

        path = Path(fpath)
        base_dir = path.parent.absolute()

        if not os.path.exists(base_dir):
            os.mkdir(base_dir)
            print(f"Created output folder {base_dir}.")

        if os.path.exists(fpath):
            logging.warning(f"Output {fpath} already exists. Overwriting it.")

        if not fpath.lower().endswith('.json'):
            logging.warning(f"Given output file {fpath} does not contain valid .json suffix.")

        with open(fpath, 'w', encoding='utf-8') as f:
            json.dump(js, f, ensure_ascii=False, indent=2)
            logging.info(f"Successfully wrote {fpath}.")
    except Exception as e:
        critical_exit(f"Writing output file {fpath} failed. {e}")


def get_utc_timestamp(time_stamp='now'):
    if time_stamp == 'now':
        dt = datetime.now(timezone.utc)

    else:
        ts = time_stamp.replace('Z', '+00:00')
        try:
            dt = datetime.fromisoformat(ts)
        except Exception as e:
            logging.error(f'invalid time stamp provided {time_stamp}: {e}.')
            SectionHandler.error_occurred = True
            return None

    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)

    return dt.isoformat(timespec='milliseconds')



import logging
from ..common.common import SectionHandler


class DocumentLeafElements(SectionHandler):
    """
    A handler for the document leaf elements which do not have any further children elements
    handling CSAF path: /document
    """
    def __init__(self, config):
        super().__init__()
        self.csaf_version = config.get('csaf_version')

    def _process_mandatory_elements(self, root_element):
        # This element is new in CSAF, not present in CVRF
        self.csaf['csaf_version'] = self.csaf_version

        self.csaf['category'] = root_element.DocumentType.text
        self.csaf['title'] = root_element.DocumentTitle.text

    def _process_optional_elements(self, root_element):
        if hasattr(root_element, 'DocumentDistribution'):
            self.csaf['distribution'] = {
                'text': root_element.DocumentDistribution.text
            }

        if hasattr(root_element, 'AggregateSeverity'):
            self.csaf['aggregate_severity'] = {
                'text': root_element.AggregateSeverity.text
            }
            if root_element.AggregateSeverity.attrib.get('Namespace'):
                self.csaf['aggregate_severity']['namespace'] = root_element.AggregateSeverity.attrib['Namespace']



import logging


class SectionHandler:
    """
    A class encapsulating arbitrary XML element in the tree.
    After conversion, it holds the JSON structure of its whole subtree.

    Attributes:
        csaf     The CSAF interpretation of the data, to be converted to JSON in the final composition step
    """

    error_occurred = False

    def __init__(self):
        self.csaf = {}

    def _process_mandatory_elements(self, root_element):
        raise NotImplementedError('Subclasses must implement.')

    def _process_optional_elements(self, root_element):
        raise NotImplementedError('Subclasses must implement.')

    def create_csaf(self, root_element):
        try:
            self._process_mandatory_elements(root_element)
        except Exception as e:
            logging.error(f'Something went wrong when processing mandatory elements for {root_element.tag}. Reason: {e}')

        try:
            self._process_optional_elements(root_element)
        except Exception as e:
            logging.error(f'Something went wrong when processing optional elements for {root_element.tag}. Reason: {e}')




import logging
import argparse
import json
import os
import re
import turvallisuusneuvonta as mandatory_tests

from lxml import etree
from lxml import objectify
from jsonschema import Draft202012Validator, ValidationError, SchemaError, draft202012_format_checker
from pkg_resources import get_distribution, Requirement, resource_filename

from .common.utils import get_config_from_file, store_json, critical_exit, create_file_name

from .section_handlers.document_leaf_elements import DocumentLeafElements
from .section_handlers.document_acknowledgments import Acknowledgments
from .section_handlers.notes import Notes
from .section_handlers.document_publisher import DocumentPublisher
from .section_handlers.references import References
from .section_handlers.document_tracking import DocumentTracking
from .section_handlers.product_tree import ProductTree
from .section_handlers.vulnerability import Vulnerability
from .common.common import SectionHandler

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(module)s - %(levelname)s - %(message)s')


class DocumentHandler:
    """
    Main Handler of the conversion:
    1. Reads/Parses/Validates CVRF XML input
    2. Iterates over each first-level XML section
    3. Each first-level XML section has its class and its methods are responsible for converting the content
    4. Collecting the output of each mapper class, which consists of the CSAF2.0 JSON equivalent
    5. Combining it to the final JSON and writing the result to a file
    """

    TOLERATED_ERRORS_SUBSTR = ["}ScoreSetV3': This element is not expected. Expected is one of ( {http://docs.oasis-open.org/csaf/ns/csaf-cvrf/v1.2/vuln}ScoreSetV2, {http://docs.oasis-open.org/csaf/ns/csaf-cvrf/v1.2/vuln}ScoreSetV3 ).",
                        "}ScoreSetV3': This element is not expected. Expected is ( {http://docs.oasis-open.org/csaf/ns/csaf-cvrf/v1.2/vuln}ScoreSetV3 ).",
                        "is not accepted by the pattern '[c][pP][eE]:/[AHOaho]?(:[A-Za-z0-9\._\-~%]*){0,6}'."]

    PACKAGE_NAME = 'cvrf2csaf'

    SCHEMA_FILE = resource_filename(Requirement.parse(PACKAGE_NAME), f'{PACKAGE_NAME}/schemata/cvrf/1.2/cvrf.xsd')
    CATALOG_FILE = resource_filename(Requirement.parse(PACKAGE_NAME), f'{PACKAGE_NAME}/schemata/catalog_1_2.xml')

    # Content copied from https://github.com/secvisogram/secvisogram/blob/main/app/lib/shared/Core/csaf_2.0_strict.json
    CSAF_SCHEMA_FILE = resource_filename(Requirement.parse(PACKAGE_NAME),
                                         f'{PACKAGE_NAME}/schemata/csaf/2.0/csaf_json_schema_strict.json')

    def __init__(self, config, pkg_version):
        self.document_leaf_elements = DocumentLeafElements(config)
        self.document_acknowledgments = Acknowledgments()
        self.document_notes = Notes()
        self.document_publisher = DocumentPublisher(config)
        self.document_references = References(config)
        self.document_tracking = DocumentTracking(config, pkg_version)
        self.product_tree = ProductTree()
        self.vulnerability = Vulnerability(config)

        self.sections_handlers = {
            'Acknowledgments': self.document_acknowledgments,
            'DocumentNotes': self.document_notes,
            'DocumentPublisher': self.document_publisher,
            'DocumentReferences': self.document_references,
            'DocumentTracking': self.document_tracking,
            'ProductTree': self.product_tree,
            'Vulnerability': self.vulnerability,
        }

    def _update_CVSSv3_version_from_schema(self, root_element):
        """ Tries to update CVSS 3.x version from schema."""
        cvss_3_regex = '.*cvss-v(3\.[01]).*'

        potential_cvss3 = None

        # iterate over  namespaces
        for ns in root_element.nsmap.values():
            match = re.match(cvss_3_regex, ns)
            if match:
                cvss_version_matched = match.groups()[0]
                # no potential cvss version found yet -> store it
                if not potential_cvss3:
                    potential_cvss3 = cvss_version_matched
                # already have some version, but it's the same as currently matched -> ok, continue
                elif potential_cvss3 == cvss_version_matched:
                    continue
                # else we have two different potential cvss versions -> skip this step comletely
                else:
                    return

        if potential_cvss3:
            logging.info(f'Default CVSS v3.x version set to {potential_cvss3} based on document XML schemas.')
            self.vulnerability.default_CVSS3_version = potential_cvss3

    def _parse(self, root):
        self._update_CVSSv3_version_from_schema(root)

        # Document leaf elements are handled on the root itself
        self.document_leaf_elements.create_csaf(root)

        # For children of the root element with a deeper structure, dedicated section handlers are used
        for elem in root.iterchildren():
            # get tag name without its namespace, don't use elem.tag here
            tag = etree.QName(elem).localname
            tag_handler = self.sections_handlers.get(tag)

            if tag_handler:
                tag_handler.create_csaf(root_element=elem)

    def _compose_final_csaf(self) -> dict:
        # Merges first level leaves into final CSAF document.
        # [mapping table](https://github.com/tschmidtb51/csaf/blob/csaf-2.0-what-is-new-table/notes/whats-new-csaf-v2.0-cn01.md#e4-mapped-elements)

        final_csaf = {'document': {}}
        final_csaf['document'] = self.document_leaf_elements.csaf

        section_mappings = (
            (final_csaf['document'], 'publisher', self.document_publisher.csaf),
            (final_csaf['document'], 'tracking', self.document_tracking.csaf),
            (final_csaf['document'], 'notes', self.document_notes.csaf),
            (final_csaf['document'], 'references', self.document_references.csaf),
            (final_csaf['document'], 'acknowledgments', self.document_acknowledgments.csaf),
            (final_csaf, 'product_tree', self.product_tree.csaf),
            (final_csaf, 'vulnerabilities', self.vulnerability.csaf),
        )

        for root, section, csaf_content in section_mappings:
            if csaf_content:
                root[section] = csaf_content

        return final_csaf

    @classmethod
    def _tolerate_errors(cls, error_list):
        tolerated_errors = [error for error in error_list if any(error_substr in error.message for error_substr in DocumentHandler.TOLERATED_ERRORS_SUBSTR)]
        if set(tolerated_errors) == set(error_list):
            logging.warning(f'Some errors during input validation happened, but can be tolerated: {tolerated_errors}.')
            return True
        return False


    @classmethod
    def _validate_input_against_schema(cls, file_path):
        with open(cls.SCHEMA_FILE) as f:
            os.environ.update(XML_CATALOG_FILES=cls.CATALOG_FILE)
            schema = etree.XMLSchema(file=f)

        try:
            schema.assertValid(file_path)
            logging.info('Input XSD validation OK.')
            return True
        except etree.DocumentInvalid as e:
            errors = list(schema.error_log)

        if not DocumentHandler._tolerate_errors(errors):
            logging.error(f'Errors during input validation occurred, reason(s): {errors}.')
            return False
        else:
            return True

    @classmethod
    def _open_and_validate_file(cls, file_path):
        try:
            xml_objectified = objectify.parse(file_path)
        except Exception as e:
            critical_exit(f'Failed to open input file {file_path}: {e}.')

        if not DocumentHandler._validate_input_against_schema(xml_objectified):
            critical_exit(f'Input document not valid, reason(s).')

        return xml_objectified.getroot()


    def convert_file(self, path) -> dict:
        """Wrapper to read/parse CVRF and parse it to CSAF JSON structure"""
        root = DocumentHandler._open_and_validate_file(path)

        self._parse(root)

        return self._compose_final_csaf()


    def validate_output_against_schema(self, final_csaf) -> bool:
        """
        Validates the CSAF output against the CSAF JSON schema
        return: True if valid, False if invalid
        """
        with open(self.CSAF_SCHEMA_FILE) as f:
            csaf_schema_content = json.loads(f.read())

        try:
            Draft202012Validator.check_schema(csaf_schema_content)
            v = Draft202012Validator(csaf_schema_content, format_checker=draft202012_format_checker)
            v.validate(final_csaf)
        except SchemaError as e:
            logging.error(f'CSAF schema validation error. Provided CSAF schema is invalid. Message: {e.message}')
            return False
        except ValidationError as e:
            logging.error(f'CSAF schema validation error. Path: {e.json_path}. Message: {e.message}.')
            return False
        else:
            logging.info('CSAF schema validation OK.')
            return True

    @staticmethod
    def validate_mandatory_tests(final_csaf):
        # TODO: After the turvallisuusneuvonta package is complete and part of the csaf package, replace the implementation
        # For now we fetch the tests like this to see which failed
        passed = True
        for m_test_str in mandatory_tests.__all__:
            # Skip translator function since translator value cannot appear on the input
            # Skip is_valid which calls all the tests (but doesnt produce any output)
            if m_test_str in ['is_valid', 'is_valid_translator']:
                continue
            m_test_call = getattr(mandatory_tests, m_test_str)
            if not m_test_call(final_csaf):
                passed = False
                logging.error(f'Mandatory test {m_test_str} failed.')

        return passed


def main():
    # General args
    parser = argparse.ArgumentParser(description='Converts CVRF 1.2 XML input into CSAF 2.0 JSON output.')
    parser.add_argument('-v', '--version', action='version', version='{}'.format(get_distribution('cvrf2csaf').version))
    parser.add_argument('--input-file', dest='input_file', type=str, required=True,
                        help="CVRF XML input file to parse", metavar='PATH')
    parser.add_argument('--output-dir', dest='output_dir', type=str, default='./', metavar='PATH',
                        help="CSAF output dir to write to. Filename is derived from /document/tracking/id.")
    parser.add_argument('--print', dest='print', action='store_true', default=False,
                        help="Additionally prints CSAF JSON output on stdout.")
    parser.add_argument('--force', action='store_const', const='cmd-arg-entered',
                        help="If used, the converter produces output even if it is invalid "
                             "(errors occured during conversion). "
                             "Target use case: best-effort conversion to JSON, "
                             "fix the errors manually, e.g. in Secvisogram.")

    # Document Publisher args
    parser.add_argument('--publisher-name', dest='publisher_name', type=str, help="Name of the publisher.")
    parser.add_argument('--publisher-namespace', dest='publisher_namespace', type=str,
                        help="Namespace of the publisher. Must be a valid URI")

    # Document Tracking args
    parser.add_argument('--fix-insert-current-version-into-revision-history', action='store_const',
                        const='cmd-arg-entered', help="If the current version is not present in the revision history "
                              "the current version is added to the revision history. Also warning is produced. "
                              "By default, an error is produced.")

    # Document References args
    parser.add_argument('--force-insert-default-reference-category', action='store_const',
                        const='cmd-arg-entered',
                        help='When "Type" attribute not present in "Reference" element, '
                             'then force using default value "external".')

    # Vulnerabilities args
    parser.add_argument('--remove-CVSS-values-without-vector', action='store_const',
                        const='cmd-arg-entered',
                        help="If vector is not present in CVSS ScoreSet, the convertor removes the whole ScoreSet "
                             "instead of producing an error. ")

    parser.add_argument('--default-CVSS3-version', dest='default_CVSS3_version',
                        help="Default version used for CVSS version 3, when the version cannot be derived from other "
                             "sources. Default value is '3.0'")


    args = {k: v for k, v in vars(parser.parse_args()).items() if v is not None}

    config = get_config_from_file()

    # Update & rewrite config file values with the ones from command line arguments
    config.update(args)

    # Boolean optional arguments that are also present in config need special treatment
    if config['fix_insert_current_version_into_revision_history'] == 'cmd-arg-entered':
        config['fix_insert_current_version_into_revision_history'] = True
    if config['force_insert_default_reference_category'] == 'cmd-arg-entered':
        config['force_insert_default_reference_category'] = True
    if config['remove_CVSS_values_without_vector'] == 'cmd-arg-entered':
        config['remove_CVSS_values_without_vector'] = True
    if config['force'] == 'cmd-arg-entered':
        config['force'] = True

    if not os.path.isfile(config.get('input_file')):
        critical_exit(f'Input file not found, check the path: {config.get("input_file")}')

    # Get the version of the installed package
    pkg_version = get_distribution('cvrf2csaf').version

    # DocumentHandler is iterating over each XML element within convert_file and return CSAF 2.0 JSON
    h = DocumentHandler(config, pkg_version)
    final_csaf = h.convert_file(path=config.get('input_file'))

    valid_output = True
    if not h.validate_output_against_schema(final_csaf) \
            or not h.validate_mandatory_tests(final_csaf) \
            or SectionHandler.error_occurred:
        valid_output = False

        if not config.get('force', False):
            critical_exit("Some error occurred during conversion, can't produce output."
                          " To override this, use --force.")
        else:
            logging.warning('Some errors occurred during conversion,'
                            ' but producing output as --force option is used.')

    # Output / Store results
    file_name = create_file_name(final_csaf['document'].get('tracking', {}).get('id', None), valid_output)
    file_path = str(os.path.join(config.get('output_dir'), file_name))
    store_json(js=final_csaf, fpath=file_path)
    if config.get('print', False):
        print(json.dumps(final_csaf, indent=2))


if __name__ == '__main__':
    main()



import logging
from ..common.common import SectionHandler


class Notes(SectionHandler):
    def __init__(self):
        super().__init__()

        self.enum_categories = {
            'description',
            'details',
            'faq',
            'general',
            'legal_disclaimer',
            'other',
            'summary'
        }

    def _process_mandatory_and_optional(self, root_element):
        notes = []

        for elem_note in root_element.Note:

            # mandatory
            new_note = dict(
                text=elem_note.text,
                category=elem_note.get('Type').lower().replace(' ', '_'),
            )

            if new_note['category'] not in self.enum_categories:
                log_msg = f'Invalid document notes category ' \
                          f'{new_note["category"]}. ' \
                          f'Should be one of: {",".join(str(x) for x in sorted(self.enum_categories))}!'
                logging.error(log_msg)
                SectionHandler.error_occurred = True

            # optional
            if elem_note.get('Audience'):
                new_note['audience'] = elem_note.get('Audience')
            if elem_note.get('Title'):
                new_note['title'] = elem_note.get('Title')

            notes.append(new_note)

        if notes and len(notes) > 0:
            self.csaf = notes

    def _process_mandatory_elements(self, root_element):
        return self._process_mandatory_and_optional(root_element=root_element)

    def _process_optional_elements(self, root_element):
        # No optional elements
        pass



import logging
from ..common.common import SectionHandler


class ProductTree(SectionHandler):
    """ Responsible for converting the ProductTree section """
    branch_type_mapping = {
        "Vendor": 'vendor',
        "Product Family": 'product_family',
        "Product Name": 'product_name',
        "Product Version": 'product_version',
        "Patch Level": 'patch_level',
        "Service Pack": 'service_pack',
        "Architecture": 'architecture',
        "Language": 'language',
        "Legacy": 'legacy',
        "Specification": 'specification',
        "Host Name": 'host_name',

        # These types do not exist in CSAF, agreed to convert it to product_name
        # https://github.com/csaf-tools/CVRF-CSAF-Converter/pull/54#discussion_r805860658
        "Realm": 'product_name',
        "Resource": 'product_name',
    }

    relation_type_mapping = {
        'Default Component Of': 'default_component_of',
        'Optional Component Of': 'optional_component_of',
        'External Component Of': 'external_component_of',
        'Installed On': 'installed_on',
        'Installed With': 'installed_with',
    }

    def __init__(self):
        super().__init__()

    def _process_mandatory_elements(self, root_element):
        """ There are no mandatory elements in the ProductTree section """
        pass

    def _process_optional_elements(self, root_element):
        self._handle_full_product_names(root_element)
        self._handle_relationships(root_element)
        self._handle_product_groups(root_element)

        branches = self._handle_branches_recursive(root_element)
        if branches is not None:
            self.csaf['branches'] = branches

    @staticmethod
    def _get_full_product_name(fpn_elem) -> dict:
        fpn = {
            'product_id': fpn_elem.attrib['ProductID'],
            'name': fpn_elem.text
        }

        if fpn_elem.attrib.get('CPE'):
            fpn['product_identification_helper'] = {'cpe': fpn_elem.attrib['CPE']}

        return fpn


    @classmethod
    def _get_branch_type(cls, branch_type: str):
        if branch_type in ['Realm', 'Resource']:
            logging.warning(f'Input branch type {branch_type} is no longer supported in CSAF. '
                            'Converting to product_name')

        return cls.branch_type_mapping[branch_type]


    def _handle_full_product_names(self, root_element):
        if not hasattr(root_element, 'FullProductName'):
            return

        full_product_names = []
        for fpn_elem in root_element.FullProductName:
            full_product_names.append(self._get_full_product_name(fpn_elem))

        self.csaf['full_product_names'] = full_product_names

    def _handle_relationships(self, root_element):
        if not hasattr(root_element, 'Relationship'):
            return

        relationships = []
        for rel_elem in root_element.Relationship:
            first_prod_name = rel_elem.FullProductName[0]

            if len(rel_elem.FullProductName) > 1:
                # To be compliant with 9.1.5 Conformance Clause 5: CVRF CSAF converter
                # https://docs.oasis-open.org/csaf/csaf/v2.0/csaf-v2.0.html
                logging.warning(f'Input line {rel_elem.sourceline}: Relationship contains more '
                                'FullProductNames. Taking only the first one, since CSAF expects '
                                'only 1 value here')

            rel_to_add = {
                'category': self.relation_type_mapping[rel_elem.attrib['RelationType']],
                'product_reference': rel_elem.attrib['ProductReference'],
                'relates_to_product_reference': rel_elem.attrib['RelatesToProductReference'],
                'full_product_name': self._get_full_product_name(first_prod_name)
            }

            relationships.append(rel_to_add)

        self.csaf['relationships'] = relationships

    def _handle_product_groups(self, root_element):
        if not hasattr(root_element, 'ProductGroups'):
            return

        product_groups = []
        for pg_elem in root_element.ProductGroups.Group:
            product_ids = [x.text for x in pg_elem.ProductID]
            pg_to_add = {
                'group_id': pg_elem.attrib['GroupID'],
                'product_ids': product_ids,
            }

            if hasattr(pg_elem, 'Description'):
                pg_to_add['summary'] = pg_elem.Description.text

            product_groups.append(pg_to_add)

        self.csaf['product_groups'] = product_groups

    def _handle_branches_recursive(self, root_element):
        """ Recursive method for handling the branches, branch can have either list of another branches,
        or a single FullProductName inside
        """
        if not hasattr(root_element, 'Branch') and not hasattr(root_element, 'FullProductName'):
            # The ProductTree section doesn't contain Branches at all
            return None

        if 'Branch' in root_element.tag and hasattr(root_element, 'FullProductName'):
            # Make sure we are inside a Branch (and not in the top ProductTree element, where FullProductName can occur)
            # then root_element is the leaf branch

            leaf_branch = {
                'name': root_element.attrib['Name'],
                'category': self._get_branch_type(root_element.attrib['Type']),
                'product': self._get_full_product_name(root_element.FullProductName)
            }

            return leaf_branch

        if hasattr(root_element, 'Branch'):
            branches = []
            for branch_elem in root_element.Branch:
                if hasattr(branch_elem, 'FullProductName'):
                    branches.append(self._handle_branches_recursive(branch_elem))
                else:
                    branches.append({
                        'name': branch_elem.attrib['Name'],
                        'category': self._get_branch_type(branch_elem.attrib['Type']),
                        'branches': self._handle_branches_recursive(branch_elem)
                    })

            return branches
